import os
import pickle


class Similarity:
    def __call__(self, query, database):
        raise NotImplementedError()

    def run_and_save(
        self, query, database, save_path, query_metadata=None, database_metadata=None
    ):
        similarities = self(query, database)

        os.makedirs(save_path, exist_ok=True)
        file_names = []
        for key, similarity in similarities.items():
            name = self.__class__.__name__ + "-" + str(key)
            data = {
                "name": key,
                "similarity": similarity,
                "metadata_query": query_metadata,
                "metadata_database": database_metadata,
            }

            file_name = os.path.join(save_path, name + ".pkl")
            with open(file_name, "wb") as file:
                pickle.dump(data, file, protocol=pickle.HIGHEST_PROTOCOL)
            file_names.append(file_name)

        return file_names



import os
from copy import deepcopy

from torch.utils.tensorboard import SummaryWriter

from wildlife_tools.tools import realize


class EpochCheckpoint:
    def __init__(self, folder=".", save_step=1):
        self.folder = folder
        self.save_step = save_step

    def __call__(self, trainer, **kwargs):
        os.makedirs(self.folder, exist_ok=True)
        if trainer.epoch % self.save_step == 0:
            trainer.save(
                folder=self.folder, file_name=f"checkpoint-{trainer.epoch}.pth"
            )


class EpochLog:
    def __init__(self, folder="."):
        self.folder = folder
        self.writer = SummaryWriter(log_dir=folder)

    def __call__(self, trainer, epoch_data: dict[str, int], **kwargs):
        if trainer.scheduler is not None:
            self.writer.add_scalar(
                "lr", trainer.scheduler.get_last_lr()[0], trainer.epoch
            )

        for key, value in epoch_data.items():
            self.writer.add_scalar(key, value, trainer.epoch)


class EpochCallbacks:
    """Returns sequence of epoch callbacks."""

    def __init__(self, steps):
        self.steps = steps

    def __call__(self, trainer, **kwargs):
        for step in self.steps:
            step(trainer=trainer, **kwargs)

    @classmethod
    def from_config(cls, config):
        config = deepcopy(config)
        steps = []
        for config_step in config["steps"]:
            steps.append(realize(config_step))
        return cls(steps=steps)



import os
import pickle


class FeatureExtractor:
    def __call__(self, dataset):
        raise NotImplementedError()

    def run_and_save(self, dataset, save_path):
        features = self(dataset)

        os.makedirs(save_path, exist_ok=True)
        name = self.__class__.__name__
        data = {
            "name": name,
            "features": features,
            "metadata": getattr(dataset, "metadata", None),
        }

        file_name = os.path.join(save_path, name + ".pkl")
        with open(file_name, "wb") as file:
            pickle.dump(data, file, protocol=pickle.HIGHEST_PROTOCOL)
        return file_name



""" Split metadata to a subset. """

from copy import deepcopy

import numpy as np

from wildlife_tools.tools import realize


class Split:
    pass


class SplitChunk(Split):
    """Returns split as given nth chunk of metadata from equally sized metadata chunks."""

    def __init__(self, chunk=1, chunk_total=1):
        self.chunk = chunk
        self.chunk_total = chunk_total

    def __call__(self, metadata):
        metadata_chunks = np.array_split(metadata, self.chunk_total)
        return metadata_chunks[self.chunk - 1]


class SplitWildlife(Split):
    """
    Returns split given wildlife splitter.
    A splitter is callable that returns list of (train_idx, test_idx) given metadata.
    Example:
        method: SplitWildlife
        splitter:
            method: closed_set
            seed: 0
            ratio_train: 0.3
        split: train
    """

    def __init__(self, splitter, split="train", split_map=None, repeat_idx=0):
        self.splitter = splitter
        self.split = split
        self.split_map = split_map
        self.repeat_idx = repeat_idx

    def __call__(self, metadata):
        if self.split_map is None:
            dataset_idx = {"train": 0, "test": 1}[self.split]
        else:
            dataset_idx = self.split_map[self.split]

        splits = self.splitter.split(metadata)
        idx = splits[self.repeat_idx][dataset_idx]
        return metadata.iloc[idx]

    @classmethod
    def from_config(cls, config):

        # TODO: add this to wildlife datasets library
        """
        class RandomProportion():
            def __init__(self, seed=1):
                self.splitter = splits.TimeProportionSplit(seed=seed)

            def split(self, df):
                final_splits = []
                for idx_train, idx_test in self.splitter.split(df):
                    final_splits.append(self.splitter.resplit_random(df, idx_train, idx_test))
                return final_splits
        """

        config = deepcopy(config)
        from wildlife_datasets import splits

        splitters = {
            "open": splits.OpenSetSplit,
            "closed": splits.ClosedSetSplit,
            "disjoint": splits.DisjointSetSplit,
            "time_proportion": splits.TimeProportionSplit,
            # "random_proportion": splits.RandomProportionSplit,
        }

        splitter_config = config.pop("splitter")
        method = splitter_config.pop("method")
        splitter = splitters[method](**splitter_config)
        return cls(splitter=splitter, **config)


class SplitMetadata(Split):
    """Returns split based on value in one of the metadata columns."""

    def __init__(self, col, value):
        self.col = col
        self.value = value

    def __call__(self, metadata):
        return metadata[metadata[self.col] == self.value]


class SplitChain(Split):
    """Returns split from sequence of splits"""

    def __init__(self, steps):
        self.steps = steps

    def __call__(self, metadata):
        for step in self.steps:
            metadata = step(metadata)
        return metadata

    @classmethod
    def from_config(cls, config):
        config = deepcopy(config)
        steps = []
        for config_step in config["steps"]:
            steps.append(realize(config_step))
        return cls(steps=steps)



import torch
import torch.nn.functional as F

from wildlife_tools.similarity.base import Similarity


class CosineSimilarity(Similarity):
    """
    Calculates cosine similarity, equivalently to `sklearn.metrics.pairwise.cosine_similarity`

    Returns:
        dict: dictionary with `cosine` key. Value is 2D array with cosine similarity.

    """

    def __call__(self, query, database):
        return {"cosine": self.cosine_similarity(query, database)}

    def cosine_similarity(self, a, b):
        a, b = torch.tensor(a), torch.tensor(b)
        similarity = torch.matmul(F.normalize(a), F.normalize(b).T)
        return similarity.numpy()




import os
import pytest
from wildlife_tools.data import WildlifeDataset, FeatureDataset
from PIL.Image import Image

load_options = ['full', 'full_mask', 'full_hide', 'bbox', 'bbox_mask', 'bbox_hide', 'crop_black']


@pytest.mark.parametrize("img_load", load_options)
def test_wildllife_dataset_img_load(metadata, img_load):
    dataset = WildlifeDataset(**metadata, img_load=img_load)
    assert len(dataset) == 3
    assert isinstance(dataset[0][0], Image)
    assert isinstance(dataset.num_classes, int)


def test_wildllife_dataset_no_label(metadata):
    dataset = WildlifeDataset(**metadata, load_label=False)
    assert len(dataset) == 3
    assert isinstance(dataset[0], Image)
    assert isinstance(dataset.num_classes, int)


def test_deep_feature_dataset(dataset, features_deep):
    feature_dataset = FeatureDataset(features_deep, metadata=dataset.metadata)
    assert len(feature_dataset) == 3
    assert all(dataset.labels_string == feature_dataset.labels_string)
    assert isinstance(dataset.num_classes, int)


def test_sift_feature_dataset_save_load(dataset, features_sift):
    a = FeatureDataset(features_sift, metadata=dataset.metadata)
    a.save('test.pkl')
    b = FeatureDataset.from_file('test.pkl')


    assert a.metadata.equals(b.metadata)
    assert len(a.features) == len(b.features)
    os.remove('test.pkl')



import timm
from wildlife_tools.train import BasicTrainer, ArcFaceLoss
from itertools import chain
import torchvision.transforms as T
from torch.optim import SGD



def test_basic_trainer(dataset_deep):
    backbone = timm.create_model('hf-hub:BVRA/MegaDescriptor-T-224', num_classes=0, pretrained=True)
    objective = ArcFaceLoss(num_classes=dataset_deep.num_classes, embedding_size=768, margin=0.5, scale=64)
    params = chain(backbone.parameters(), objective.parameters())
    optimizer = SGD(params=params, lr=0.001, momentum=0.9)

    trainer = BasicTrainer(
        dataset=dataset_deep,
        model=backbone,
        objective=objective,
        optimizer=optimizer,
        epochs=2,
        device='cpu',
    )
    trainer.train()




import json
import os
import pickle
from typing import Callable

import cv2
import numpy as np
import pandas as pd
import pycocotools.mask as mask_coco
from PIL import Image

from wildlife_tools.tools import realize


class WildlifeDataset:
    """
    PyTorch-style dataset for a wildlife datasets

    Args:
        metadata: A pandas dataframe containing image metadata.
        root: Root directory if paths in metadata are relative. If None, paths in metadata are used.
        split: A function that splits metadata, e.g., instance of data.Split.
        transform: A function that takes in an image and returns its transformed version.
        img_load: Method to load images.
            Options: 'full', 'full_mask', 'full_hide', 'bbox', 'bbox_mask', 'bbox_hide',
                      and 'crop_black'.
        col_path: Column name in the metadata containing image file paths.
        col_label: Column name in the metadata containing class labels.
        load_label: If False, \_\_getitem\_\_ returns only image instead of (image, label) tuple.

    Attributes:
        labels np.array : An integers array of ordinal encoding of labels.
        labels_string np.array: A strings array of original labels.
        labels_map dict: A mapping between labels and their ordinal encoding.
        num_classes int: Return the number of unique classes in the dataset.
    """

    def __init__(
        self,
        metadata: pd.DataFrame,
        root: str | None = None,
        split: Callable | None = None,
        transform: Callable | None = None,
        img_load: str = "full",
        col_path: str = "path",
        col_label: str = "identity",
        load_label: bool = True,
    ):
        self.split = split
        if self.split:
            metadata = self.split(metadata)

        self.metadata = metadata.reset_index(drop=True)
        self.root = root
        self.transform = transform
        self.img_load = img_load
        self.col_path = col_path
        self.col_label = col_label
        self.load_label = load_label
        self.labels, self.labels_map = pd.factorize(
            self.metadata[self.col_label].values
        )

    @property
    def labels_string(self):
        return self.metadata[self.col_label].astype(str).values

    @property
    def num_classes(self):
        return len(self.labels_map)

    def __len__(self):
        return len(self.metadata)

    def get_image(self, path):
        img = cv2.imread(path)
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        img = Image.fromarray(img)
        return img

    def __getitem__(self, idx):
        data = self.metadata.iloc[idx]
        if self.root:
            img_path = os.path.join(self.root, data[self.col_path])
        else:
            img_path = data[self.col_path]
        img = self.get_image(img_path)

        if self.img_load in ["full_mask", "full_hide", "bbox_mask", "bbox_hide"]:
            if not ("segmentation" in data):
                raise ValueError(f"{self.img_load} selected but no segmentation found.")
            if type(data["segmentation"]) == str:
                segmentation = eval(data["segmentation"])
            else:
                segmentation = data["segmentation"]

        if self.img_load in ["bbox", "bbox_mask", "bbox_hide"]:
            if not ("bbox" in data):
                raise ValueError(f"{self.img_load} selected but no bbox found.")
            if type(data["bbox"]) == str:
                bbox = json.loads(data["bbox"])
            else:
                bbox = data["bbox"]

        # Load full image as it is.
        if self.img_load == "full":
            img = img

        # Mask background using segmentation mask.
        elif self.img_load == "full_mask":
            mask = mask_coco.decode(segmentation).astype("bool")
            img = Image.fromarray(img * mask[..., np.newaxis])

        # Hide object using segmentation mask
        elif self.img_load == "full_hide":
            mask = mask_coco.decode(segmentation).astype("bool")
            img = Image.fromarray(img * ~mask[..., np.newaxis])

        # Crop to bounding box
        elif self.img_load == "bbox":
            img = img.crop((bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]))

        # Mask background using segmentation mask and crop to bounding box.
        elif self.img_load == "bbox_mask":
            mask = mask_coco.decode(segmentation).astype("bool")
            img = Image.fromarray(img * mask[..., np.newaxis])
            img = img.crop((bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]))

        # Hide object using segmentation mask and crop to bounding box.
        elif self.img_load == "bbox_hide":
            mask = mask_coco.decode(segmentation).astype("bool")
            img = Image.fromarray(img * ~mask[..., np.newaxis])
            img = img.crop((bbox[0], bbox[1], bbox[0] + bbox[2], bbox[1] + bbox[3]))

        # Crop black background around images
        elif self.img_load == "crop_black":
            y_nonzero, x_nonzero, _ = np.nonzero(img)
            img = img.crop(
                (
                    np.min(x_nonzero),
                    np.min(y_nonzero),
                    np.max(x_nonzero),
                    np.max(y_nonzero),
                )
            )

        else:
            raise ValueError(f"Invalid img_load argument: {self.img_load}")

        if self.transform:
            img = self.transform(img)

        if self.load_label:
            return img, self.labels[idx]
        else:
            return img

    @classmethod
    def from_config(cls, config):
        config["split"] = realize(config.get("split"))
        config["transform"] = realize(config.get("transform"))
        config["metadata"] = pd.read_csv(config["metadata"], index_col=False)
        return cls(**config)


class FeatureDataset:
    def __init__(
        self,
        features,
        metadata,
        col_label="identity",
        load_label=True,
    ):

        if len(features) != len(metadata):
            raise ValueError("Features and metadata (lables) have different length ! ")

        self.load_label = load_label
        self.features = features
        self.metadata = metadata.reset_index(drop=True)
        self.col_label = col_label
        self.labels, self.labels_map = pd.factorize(
            self.metadata[self.col_label].values
        )

    @property
    def labels_string(self):
        return self.metadata[self.col_label].astype(str).values

    def __getitem__(self, idx):
        if self.load_label:
            return self.features[idx], self.labels[idx]
        else:
            return self.features[idx]

    def __len__(self):
        return len(self.metadata)

    @property
    def num_classes(self):
        return len(self.labels_map)

    def save(self, path):
        data = {
            "features": self.features,
            "metadata": self.metadata,
            "col_label": self.col_label,
            "load_label": self.load_label,
        }
        dirname = os.path.dirname(path)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(path, "wb") as file:
            pickle.dump(data, file, protocol=pickle.HIGHEST_PROTOCOL)

    @classmethod
    def from_file(cls, path, **config):
        with open(path, "rb") as file:
            data = pickle.load(file)
        return cls(**data, **config)

    @classmethod
    def from_config(cls, config):
        path = config.pop("path")
        return cls.load(path, **config)


class FeatureDatabase(FeatureDataset):
    """Alias for FeatureDataset"""

    pass



import cv2
import numpy as np
from tqdm import tqdm

from wildlife_tools.data import WildlifeDataset
from wildlife_tools.features.base import FeatureExtractor


class SIFTFeatures(FeatureExtractor):
    """
    Extracts SIFT descriptors for each image in the dataset.

    Args:
        max_keypoints: Limit number of extracted keypoints / descriptors.

    Returns:
        list of arrays, each corresponds to an input image and have shape `[n_descriptors x 128]`.
    """

    descriptor_dim: int = 128

    def __init__(self, max_keypoints: int | None = None):
        self.max_keypoints = max_keypoints

    def __call__(self, dataset: WildlifeDataset):
        if self.max_keypoints:
            sift = cv2.SIFT_create(nfeatures=self.max_keypoints)
        else:
            sift = cv2.SIFT_create()

        descriptors = []
        for img, y in tqdm(dataset, mininterval=1, ncols=100):
            keypoint, d = sift.detectAndCompute(np.array(img), None)
            if len(keypoint) <= 1:
                descriptors.append(None)
            else:
                descriptors.append(d)
        return descriptors



from .dataset import FeatureDatabase, FeatureDataset, WildlifeDataset
from .split import SplitChain, SplitChunk, SplitMetadata, SplitWildlife
from .transform import TransformTimm, TransformTorchvision



import torch
from tqdm import tqdm

from wildlife_tools.features.base import FeatureExtractor
from wildlife_tools.features.models.superpoint import SuperPoint


class SuperPointFeatures(FeatureExtractor):
    def __init__(
        self,
        descriptor_dim: int = 256,
        max_keypoints: int | None = None,
        nms_radius: int = 4,
        keypoint_threshold: float = 0.005,
        remove_borders: int = 4,
        device: str = "cpu",
        num_workers: int = 1,
        batch_size: int = 128,
    ):
        self.descriptor_dim = descriptor_dim
        self.max_keypoints = max_keypoints
        self.nms_radius = nms_radius
        self.keypoint_threshold = keypoint_threshold
        self.remove_borders = remove_borders
        self.device = device
        self.num_workers = num_workers
        self.batch_size = batch_size

    def __call__(self, dataset):
        if not self.max_keypoints:
            max_keypoints = -1
        else:
            max_keypoints = self.max_keypoints

        model = SuperPoint(
            config={
                "descriptor_dim": self.descriptor_dim,
                "nms_radius": self.nms_radius,
                "keypoint_threshold": self.keypoint_threshold,
                "max_keypoints": max_keypoints,
                "remove_borders": self.remove_borders,
            }
        ).to(self.device)
        loader = torch.utils.data.DataLoader(
            dataset,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
            shuffle=False,
        )

        descriptors = []
        for image, label in tqdm(loader, mininterval=1, ncols=100):
            with torch.no_grad():
                output = model({"image": image.to(self.device)})
            descriptors.extend(
                [d.permute(1, 0).cpu().numpy() for d in output["descriptors"]]
            )
        return descriptors



import itertools

import kornia.feature as KF
import numpy as np
import pandas as pd
import torch
from tqdm import tqdm

from wildlife_tools.similarity.base import Similarity


def batched(iterable, n):
    """
    Batch data into tuples of length n. The last batch may be shorter.
    Example: batched('ABCDEFG', 3) --> ABC DEF G
    """
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(itertools.islice(it, n)):
        yield batch


class MatchLOFTR(Similarity):
    """
    Calculate similarity between query and database as number of descriptors correspondences
    after filtering with Low ratio test.

    Args:
        pretrained: LOFTR model used. `outdoor` or `indoor`.
        thresholds: Iterable with confidence thresholds. Should be in [0, 1] interval.
        batch_size: Batch size used for the inference.
        device: Specifies device used for the inference.
        silent: disable tqdm bar.

    Returns:
        dict: Values are 2D array with number of correspondences for each threshold.
    """

    def __init__(
        self,
        pretrained: str = "outdoor",
        thresholds: tuple[float] = (0.99,),
        batch_size: int = 128,
        device: str = "cuda",
        silent: bool = False,
    ):
        self.device = device
        self.matcher = KF.LoFTR(pretrained=pretrained).to(device)
        self.thresholds = thresholds
        self.batch_size = batch_size
        self.silent = silent

    def __call__(self, query, database):
        iterator = batched(
            itertools.product(enumerate(query), enumerate(database)), self.batch_size
        )
        iterator_size = int(np.ceil(len(query) * len(database) / self.batch_size))
        similarities = {
            t: np.full((len(query), len(database)), np.nan, dtype=np.float16)
            for t in self.thresholds
        }

        for pair_batch in tqdm(
            iterator, total=iterator_size, mininterval=1, ncols=100, disable=self.silent
        ):
            q, d = zip(*pair_batch)
            q_idx, q_data = list(zip(*q))
            d_idx, d_data = list(zip(*d))
            input_dict = {
                "image0": torch.stack(q_data).to(self.device),
                "image1": torch.stack(d_data).to(self.device),
            }
            with torch.inference_mode():
                correspondences = self.matcher(input_dict)

            batch_idx = correspondences["batch_indexes"].cpu().numpy()
            confidence = correspondences["confidence"].cpu().numpy()
            for t in self.thresholds:
                series = pd.Series(confidence > t)
                for j, group in series.groupby(batch_idx):
                    similarities[t][q_idx[j], d_idx[j]] = group.sum()
        return similarities



from tqdm import tqdm

from wildlife_tools.data import WildlifeDataset
from wildlife_tools.features.base import FeatureExtractor


class DataToMemory(FeatureExtractor):
    """Loads dataset to memory for faster access."""

    def __call__(self, dataset: WildlifeDataset):
        features = []
        for x, y in tqdm(dataset, mininterval=1, ncols=100):
            features.append(x)
        return features



import torch.nn as nn
from pytorch_metric_learning import distances, losses, miners


class ArcFaceLoss(nn.Module):
    """
    Wraps Pytorch Metric Learning ArcFaceLoss.

    Default margin (in radians): 0.5
    Default scale: 64
    """

    def __init__(self, num_classes, embedding_size, margin=0.5, scale=64):
        super().__init__()
        self.loss = losses.ArcFaceLoss(
            num_classes=num_classes,
            embedding_size=embedding_size,
            margin=57.3 * margin,
            scale=scale,
        )

    def forward(self, embeddings, y):
        return self.loss(embeddings, y)


class TripletLoss(nn.Module):
    """
    Wraps Pytorch Metric Learning TripletMarginLoss.

    Mining is one of: 'all', 'hard', 'semihard'
    Distance is one of: 'cosine', 'l2', 'l2_squared'
    """

    def __init__(self, margin=0.2, mining="seminard", distance="l2_squared"):
        super().__init__()
        if distance == "cosine":
            distance = distances.CosineSimilarity()
        elif distance == "l2":
            distance = distances.LpDistance(normalize_embeddings=True, p=2, power=1)
        elif distance == "l2_squared":
            distance = distances.LpDistance(normalize_embeddings=True, p=2, power=2)
        else:
            raise ValueError(f"Invalid distance: {distance}")

        self.loss = losses.TripletMarginLoss(distance=distance, margin=margin)
        self.miner = miners.TripletMarginMiner(
            distance=distance, type_of_triplets=mining, margin=margin
        )

    def forward(self, embeddings, y):
        indices_tuple = self.miner(embeddings, y)
        return self.loss(embeddings, y, indices_tuple)


class SoftmaxLoss(nn.Module):
    """CE with single dense layer classification head."""

    def __init__(self, num_classes, embedding_size):
        super().__init__()
        self.criterion = nn.CrossEntropyLoss()
        self.linear = nn.Linear(embedding_size, num_classes)

    def forward(self, x, y):
        return self.criterion(self.linear(x), y)



import timm
from wildlife_tools.features import SIFTFeatures, DeepFeatures, DataToMemory


def test_features_deep(dataset_deep):
    backbone = timm.create_model('hf-hub:BVRA/MegaDescriptor-T-224', num_classes=0, pretrained=True)
    extractor = DeepFeatures(backbone)
    output = extractor(dataset_deep)
    assert len(output) == len(dataset_deep)
    assert output.shape[1] == 768


def test_features_sift(dataset_sift):
    extractor = SIFTFeatures()
    output = extractor(dataset_sift)
    print(output)
    assert len(output) == len(dataset_sift)
    assert output[0].shape[1] == 128


def test_data_memory(dataset_deep):
    extractor = DataToMemory()
    output = extractor(dataset_deep)
    assert len(output) == len(dataset_deep)



from .backbone import TimmBackbone
from .callbacks import EpochCallbacks
from .objective import ArcFaceLoss, SoftmaxLoss, TripletLoss
from .optim import OptimizerAdam, OptimizerSGD
from .trainer import (BasicTrainer, ClassifierTrainer, EmbeddingTrainer,
                      set_seed)



import os
import random
from copy import deepcopy
from itertools import chain

import numpy as np
import torch
import torch.backends.cudnn
from tqdm import tqdm

from wildlife_tools.tools import realize


def set_seed(seed=0):
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def set_random_states(states):
    if 'os_rng_state' in states and states["os_rng_state"]:
        os.environ["PYTHONHASHSEED"] = states["os_rng_state"]
    if 'random_rng_state' in states:
        random.setstate(states["random_rng_state"])
    if 'numpy_rng_state' in states:
        np.random.set_state(states["numpy_rng_state"])
    if 'torch_rng_state' in states:
        torch.set_rng_state(states["torch_rng_state"])
    if 'torch_cuda_rng_state' in states:
        torch.cuda.set_rng_state(states["torch_cuda_rng_state"])
    if 'torch_cuda_rng_state_all' in states:
        torch.cuda.set_rng_state_all(states["torch_cuda_rng_state_all"])

    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False


def get_random_states():
    states = {}
    states["os_rng_state"] = os.environ.get("PYTHONHASHSEED")
    states["random_rng_state"] = random.getstate()
    states["numpy_rng_state"] = np.random.get_state()
    states["torch_rng_state"] = torch.get_rng_state()
    if torch.cuda.is_available():
        states["torch_cuda_rng_state"] = torch.cuda.get_rng_state()
        states["torch_cuda_rng_state_all"] = torch.cuda.get_rng_state_all()
    return states


class BasicTrainer:
    def __init__(
        self,
        dataset,
        model,
        objective,
        optimizer,
        epochs,
        scheduler=None,
        device="cuda",
        batch_size=128,
        num_workers=1,
        accumulation_steps=1,
        epoch_callback=None,
    ):
        self.dataset = dataset
        self.model = model.to(device)
        self.objective = objective.to(device)
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.epochs = epochs
        self.epoch = 0
        self.device = device
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.accumulation_steps = accumulation_steps
        self.epoch_callback = epoch_callback

    def train(self):
        loader = torch.utils.data.DataLoader(
            self.dataset,
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=True,
        )

        for e in range(self.epochs):
            epoch_data = self.train_epoch(loader)
            self.epoch += 1

            if self.epoch_callback:
                self.epoch_callback(trainer=self, epoch_data=epoch_data)

    def train_epoch(self, loader):
        model = self.model.train()
        losses = []
        for i, batch in enumerate(
            tqdm(loader, desc=f"Epoch {self.epoch}: ", mininterval=1, ncols=100)
        ):
            x, y = batch
            x, y = x.to(self.device), y.to(self.device)

            out = model(x)
            loss = self.objective(out, y)
            loss.backward()
            if (i - 1) % self.accumulation_steps == 0:
                self.optimizer.step()
                self.optimizer.zero_grad()

            losses.append(loss.detach().cpu())

        if self.scheduler:
            self.scheduler.step()

        return {"train_loss_epoch_avg": np.mean(losses)}

    def save(self, folder, file_name="checkpoint.pth", save_rng=True, **kwargs):
        if not os.path.exists(folder):
            os.makedirs(folder)

        checkpoint = {}
        checkpoint["model"] = self.model.state_dict()
        checkpoint["objective"] = self.objective.state_dict()
        checkpoint["optimizer"] = self.optimizer.state_dict()
        checkpoint["epoch"] = self.epoch
        if save_rng:
            checkpoint["rng_states"] = get_random_states()
        if self.scheduler:
            checkpoint["scheduler"] = self.scheduler.state_dict()

        torch.save(checkpoint, os.path.join(folder, file_name))

    def load(self, path, load_rng=True):
        checkpoint = torch.load(path, map_location=torch.device(self.device))

        if "model" in checkpoint:
            self.model.load_state_dict(checkpoint["model"])
        if "objective" in checkpoint:
            self.objective.load_state_dict(checkpoint["objective"])
        if "optimizer" in checkpoint:
            self.optimizer.load_state_dict(checkpoint["optimizer"])
        if "epoch" in checkpoint:
            self.epoch = checkpoint["epoch"]
        if "rng_states" in checkpoint and load_rng:
            set_random_states(checkpoint["rng_states"])
        if "scheduler" in checkpoint and self.scheduler:
            self.scheduler.load_state_dict(checkpoint["scheduler"])


class ClassifierTrainer:
    @classmethod
    def from_config(cls, config):
        """
        Use config dict to setup BasicTrainer for training classifier.

        Config keys:
            dataset (dict):
                Config dictionary of the training dataset.
            backbone (dict):
                Config dictionary of the backbone.
            objective (dict):
                Config dictionary of the objective.
            scheduler (dict | None, default: None):
                Config dictionary of the scheduler (no scheduler is used by default).
            epochs (int):
                Number of training epochs.
            device (str, default: 'cuda'):
                Device to be used.
            batch_size (int, default: 128):
                Training batch size.
            num_workers (int, default: 1):
                Number of data loading workers in torch DataLoader.
            accumulation_steps (int, default: 1):
                Number of gradient accumulation steps.

        Returns:
            Ready to use BasicTrainer

        """

        config = deepcopy(config)

        dataset = realize(
            config=config.pop("dataset"),
        )
        model = realize(
            config=config.pop("backbone"),
            output_size=dataset.num_classes,
        )
        objective = realize(
            config=config.pop("objective"),
        )
        optimizer = realize(
            config=config.pop("optimizer"),
            params=model.parameters(),
        )
        scheduler = realize(
            config=config.pop("scheduler", None),
            epochs=config.get("epochs"),
        )
        epoch_callback = realize(
            config=config.pop("epoch_callback", None),
        )

        return BasicTrainer(
            model=model,
            dataset=dataset,
            objective=objective,
            optimizer=optimizer,
            scheduler=scheduler,
            epoch_callback=epoch_callback,
            **config,
        )


class EmbeddingTrainer:
    @classmethod
    def from_config(cls, config):
        """Use config dict to setup BasicTrainer for training embedder.

        Config keys:
            dataset (dict):
                Config dictionary of the training dataset.
            backbone (dict):
                Config dictionary of the backbone.
            objective (dict):
                Config dictionary of the objective.
            scheduler (dict | None, default: None):
                Config dictionary of the scheduler (no scheduler is used by default).
            embedding_size (int | None, default: None):
                Adds a linear layer after the backbone with the target embedding size.
                By default, embedding size is inferred from backbone (e.g., num_classes=0 in TIMM).
            epochs (int):
                Number of training epochs.
            device (str, default: 'cuda'):
                Device to be used.
            batch_size (int, default: 128):
                Training batch size.
            num_workers (int, default: 1):
                Number of data loading workers in torch DataLoader.
            accumulation_steps (int, default: 1):
                Number of gradient accumulation steps.

        Returns:
            Instance of BasicTrainer

        """

        config = deepcopy(config)
        embedding_size = config.pop("embedding_size", None)

        dataset = realize(config=config.pop("dataset"))
        backbone = realize(config=config.pop("backbone"), output_size=embedding_size)

        if embedding_size is None:  # Infer embedding size
            with torch.no_grad():
                x = dataset[0][0].unsqueeze(0)
                embedding_size = backbone(x).shape[1]

        objective = realize(
            config=config.pop("objective"),
            embedding_size=embedding_size,
            num_classes=dataset.num_classes,
        )
        optimizer = realize(
            config=config.pop("optimizer"),
            params=chain(backbone.parameters(), objective.parameters()),
        )
        scheduler = realize(
            optimizer=optimizer,
            config=config.pop("scheduler", None),
            epochs=config.get("epochs"),
        )
        epoch_callback = realize(
            config=config.pop("epoch_callback", None),
        )
        return BasicTrainer(
            model=backbone,
            dataset=dataset,
            objective=objective,
            optimizer=optimizer,
            scheduler=scheduler,
            epoch_callback=epoch_callback,
            **config,
        )



import importlib
import os
import random
from copy import deepcopy

import numpy as np
import torch
import torch.backends.cudnn
import yaml
from jinja2 import Environment, meta


class Store:
    """Dictionary with components. Components are imported as needed."""

    modules = {
        # Data
        "SplitChunk": ("wildlife_tools.data.split", "SplitChunk"),
        "SplitWildlife": ("wildlife_tools.data.split", "SplitWildlife"),
        "SplitMetadata": ("wildlife_tools.data.split", "SplitMetadata"),
        "SplitChain": ("wildlife_tools.data.split", "SplitChain"),
        "TransformTimm": ("wildlife_tools.data.transform", "TransformTimm"),
        "TransformTorchvision": (
            "wildlife_tools.data.transform",
            "TransformTorchvision",
        ),
        "WildlifeDataset": ("wildlife_tools.data.dataset", "WildlifeDataset"),
        "FeatureDataset": ("wildlife_tools.data.dataset", "FeatureDataset"),
        # Features & Similarity
        "SimilarityPipeline": (
            "wildlife_tools.pipelines.similarity",
            "SimilarityPipeline",
        ),
        # Train
        "TrainingPipeline": ("wildlife_tools.pipelines.train", "TrainingPipeline"),
        "EpochCallbacks": ("wildlife_tools.train.callbacks", "EpochCallbacks"),
        "EpochLog": ("wildlife_tools.train.callbacks", "EpochLog"),
        "EpochCheckpoint": ("wildlife_tools.train.callbacks", "EpochCheckpoint"),
        "EmbeddingTrainer": ("wildlife_tools.train.trainer", "EmbeddingTrainer"),
        "ClassifierTrainer": ("wildlife_tools.train.trainer", "ClassifierTrainer"),
        "OptimizerAdam": ("wildlife_tools.train.optim", "OptimizerAdam"),
        "OptimizerSGD": ("wildlife_tools.train.optim", "OptimizerSGD"),
        "SchedulerCosine": ("wildlife_tools.train.optim", "SchedulerCosine"),
        "TimmBackbone": ("wildlife_tools.train.backbone", "TimmBackbone"),
        "ArcFaceLoss": ("wildlife_tools.train.objective", "ArcFaceLoss"),
        "TripletLoss": ("wildlife_tools.train.objective", "TripletLoss"),
        "SoftmaxLoss": ("wildlife_tools.train.objective", "SoftmaxLoss"),
    }

    def __getitem__(self, key):
        module_path, class_name = self.modules[key]
        module = importlib.import_module(module_path)
        return getattr(module, class_name)

    def __contains__(self, key):
        return key in self.modules


def realize(config, store=None, **kwargs):
    """realize object from config given its name and method store"""

    if config is None:
        return None

    if store is None:
        store = Store()

    config_local = deepcopy(config)
    name = config_local.pop("method", None)

    if name is None:
        raise ValueError("No 'method' attribute in config")
    if name not in store:
        raise ValueError(f"No method {name} in method store.")

    obj_class = store[name]
    if hasattr(obj_class, "from_config"):
        obj = obj_class.from_config(config=config_local, **kwargs)
    else:
        obj = obj_class(**config_local, **kwargs)

    obj.source_config = deepcopy(config)
    return obj


def parse_yaml(yaml_string):
    '''
    Impute variables with in "{{ }}" with values from top level of the yaml dictionary.
    Example:

    yaml_string = """
    a:
        name: test
    b:
        uses: "{{a}}"
    """
    parse_yaml(yaml_string)
    >>> {'a': {'name': 'test'}, 'b': {'uses': {'name': 'test'}}}
    '''

    env = Environment(
        variable_start_string="'{{",
        variable_end_string="}}'",
    )

    data = yaml.safe_load(yaml_string)
    data_str = str(data)
    variables = meta.find_undeclared_variables(env.parse(data_str))

    while len(variables) != 0:
        data_str = env.from_string(data_str).render(data)
        data = yaml.safe_load(data_str)
        variables_new = meta.find_undeclared_variables(env.parse(data_str))

        # Check if number of resolved variables changed
        if len(variables) == len(variables_new):
            raise ValueError(f"Unable to impute variables in all places: {variables}")

        variables = variables_new
    return data


def set_seed(seed=0, device="cuda"):
    os.environ["PYTHONHASHSEED"] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    if device == "cuda":
        torch.cuda.manual_seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False



from __future__ import annotations

from collections import defaultdict

import numpy as np
import pandas as pd
import torch

from wildlife_tools.similarity import CosineSimilarity


class KnnClassifier:
    """
    Predict query label as k labels of nearest matches in database. If there is tie at given k,
    prediction from k-1 is used. Input is similarity matrix with `n_query` x `n_database` shape.


    Args:
        k: use k nearest in database for the majority voting.
        database_labels: list of labels in database. If provided, decode predictions to database
        (e.g. string) labels.
    Returns:
        1D array with length `n_query` of database labels (col index of the similarity matrix).
    """

    def __init__(self, k: int = 1, database_labels: np.array | None = None):
        self.k = k
        self.database_labels = database_labels

    def __call__(self, similarity):
        similarity = torch.tensor(similarity, dtype=float)
        scores, idx = similarity.topk(k=self.k, dim=1)
        pred = self.aggregate(idx)[:, self.k - 1]

        if self.database_labels is not None:
            pred = self.database_labels[pred]
        return pred

    def aggregate(self, predictions):
        """
        Aggregates array of nearest neighbours to single prediction for each k.
        If there is tie at given k, prediction from k-1 is used.

        Args:
            array of with shape [n_query, k] of nearest neighbours.
        Returns:
            array with predictions [n_query, k]. Column dimensions are predictions for [k=1,...,k=k]
        """

        results = defaultdict(list)
        for k in range(1, predictions.shape[1] + 1):
            for row in predictions[:, :k]:
                vals, counts = np.unique(row, return_counts=True)
                best = vals[np.argmax(counts)]

                counts_sorted = sorted(counts)
                if (len(counts_sorted)) > 1 and (counts_sorted[0] == counts_sorted[1]):
                    best = None
                results[k].append(best)

        results = pd.DataFrame(results).T.fillna(method="ffill").T
        return results.values


class KnnMatcher:
    """
    Find nearest match to query in existing database of features.
    Combines CosineSimilarity and KnnClassifier.
    """

    def __init__(self, database, k=1):
        self.similarity = CosineSimilarity()
        self.database = database
        self.classifier = KnnClassifier(
            database_labels=self.database.labels_string, k=k
        )

    def __call__(self, query):
        if isinstance(query, list):
            query = np.concatenate(query)

        if not isinstance(query, np.ndarray):
            raise ValueError("Query should be array or list of features.")

        sim_matrix = self.similarity(query, self.database.features)["cosine"]
        return self.classifier(sim_matrix)




from wildlife_tools.similarity import CosineSimilarity, MatchDescriptors, MatchLOFTR


def test_cosine_similarity(features_deep):
    method = CosineSimilarity()
    output = method(features_deep, features_deep)
    assert 'cosine' in output
    assert output['cosine'].shape == (3, 3)


def test_match_descriptors(features_sift):
    similarity = MatchDescriptors(descriptor_dim=128, thresholds=[0.8])
    output = similarity(features_sift, features_sift)
    assert 0.8 in output
    assert output[0.8].shape == (3, 3)


def test_match_loftr(dataset_loftr):
    similarity = MatchLOFTR(device='cpu', thresholds=[0.8])
    output = similarity(dataset_loftr, dataset_loftr)
    assert 0.8 in output
    assert output[0.8].shape == (3, 3)


import torch
from tqdm import tqdm
from transformers import CLIPModel, CLIPProcessor

from wildlife_tools.data import WildlifeDataset
from wildlife_tools.features.base import FeatureExtractor
from wildlife_tools.tools import realize


class DeepFeatures(FeatureExtractor):
    """
    Extracts features using forward pass of pytorch model.

    Args:
        model: Pytorch model used for the feature extraction.
        batch_size: Batch size used for the feature extraction.
        num_workers: Number of workers used for data loading.
        device: Select between cuda and cpu devices.

    Returns:
        An array with a shape of `n_input` x `dim_embedding`.

    """

    def __init__(
        self,
        model,
        batch_size: int = 128,
        num_workers: int = 1,
        device: str = "cpu",
    ):
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.device = device
        self.model = model

    def __call__(self, dataset: WildlifeDataset):
        self.model = self.model.to(self.device)
        self.model = self.model.eval()

        loader = torch.utils.data.DataLoader(
            dataset,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
            shuffle=False,
        )
        outputs = []
        for image, label in tqdm(loader, mininterval=1, ncols=100):
            with torch.no_grad():
                output = self.model(image.to(self.device))
                outputs.append(output.cpu())
        return torch.cat(outputs).numpy()

    @classmethod
    def from_config(cls, config):
        model = realize(config.pop("model"))
        return cls(model=model, **config)


class ClipFeatures:
    """
    Extract features using CLIP model (https://arxiv.org/pdf/2103.00020.pdf).
    Uses raw images of input WildlifeDataset (i.e. dataset.transform = None)

    Args:
        model: transformer.CLIPModel. Uses VIT-L backbone by default.
        processor: transformer.CLIPProcessor. Uses VIT-L processor by default.
        batch_size: Batch size used for the feature extraction.
        num_workers: Number of workers used for data loading.
        device: Select between cuda and cpu devices.

    Returns:
        An array with a shape of `n_input` x `dim_embedding`.
    """

    def __init__(
        self,
        model=None,
        processor=None,
        batch_size=128,
        num_workers=1,
        device="cpu",
    ):
        if model is None:
            model = CLIPModel.from_pretrained(
                "openai/clip-vit-large-patch14"
            ).vision_model

        if processor is None:
            processor = CLIPProcessor.from_pretrained("openai/clip-vit-large-patch14")

        self.model = model
        self.processor = processor
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.device = device
        self.transform = lambda x: processor(images=x, return_tensors="pt")[
            "pixel_values"
        ]

    def __call__(self, dataset: WildlifeDataset):
        self.model = self.model.to(self.device)
        self.model = self.model.eval()

        dataset.transforms = None  # Reset transforms.

        loader = torch.utils.data.DataLoader(
            dataset,
            num_workers=self.num_workers,
            batch_size=self.batch_size,
            shuffle=False,
            collate_fn=lambda x: x,
        )
        outputs = []
        for image in tqdm(loader, mininterval=1, ncols=100):
            with torch.no_grad():
                output = self.model(self.transform(image).to(self.device)).pooler_output
                outputs.append(output.cpu())
        return torch.cat(outputs).numpy()




from wildlife_tools.inference import KnnMatcher, KnnClassifier
from wildlife_tools.data import FeatureDataset


def test_knn_classifier_deep(similarity_deep, dataset):
    classifier = KnnClassifier(k=1, database_labels=dataset.labels_string)
    output = classifier(similarity_deep)
    assert len(output) == len(dataset)


def test_knn_classifier_sift(similarity_sift, dataset):
    classifier = KnnClassifier(k=1, database_labels=dataset.labels_string)
    output = classifier(similarity_sift)
    assert len(output) == len(dataset)


def test_knn_matcher(features_deep, dataset):
    feature_dataset = FeatureDataset(features_deep, metadata=dataset.metadata)
    matcher = KnnMatcher(feature_dataset)
    output = matcher(features_deep)
    assert len(output) == len(features_deep)


